@@ -6,11 +6,11 @@
 // copied, modified, or distributed except according to those terms.
 //
 
-use crate::{read::LineDecoder, request, response, Backend};
+use crate::{read::LineDecoder, request, response, Backend, BoxStream};
 use async_trait::async_trait;
 use bytes::Bytes;
 use common_multipart_rfc7578::client::multipart;
-use futures::{future, FutureExt, Stream, TryStreamExt};
+use futures::{future, FutureExt, TryStreamExt};
 use std::{
     fs::File,
     io::{Cursor, Read},
@@ -33,14 +33,15 @@ macro_rules! impl_stream_api_response {
         }
     };
     (($self:ident, $req:expr, $form:expr) |$var:ident| => $impl:block) => {
-        match $self.build_base_request(&$req, $form) {
+        match $self.build_base_request($req, $form) {
             Ok($var) => $impl,
             Err(e) => Box::new(future::err(e).into_stream()),
         }
     };
 }
 
-#[async_trait(?Send)]
+#[cfg_attr(feature = "with-send-sync", async_trait)]
+#[cfg_attr(not(feature = "with-send-sync"), async_trait(?Send))]
 pub trait IpfsApi: Backend {
     /// Add file to Ipfs.
     ///
@@ -57,7 +58,7 @@ pub trait IpfsApi: Backend {
     ///
     async fn add<R>(&self, data: R) -> Result<response::AddResponse, Self::Error>
     where
-        R: 'static + Read + Send + Sync,
+        R: 'static + Read + Send + Sync + Unpin,
     {
         self.add_with_options(data, request::Add::default()).await
     }
@@ -94,7 +95,7 @@ pub trait IpfsApi: Backend {
         add: request::Add<'_>,
     ) -> Result<response::AddResponse, Self::Error>
     where
-        R: 'static + Read + Send + Sync,
+        R: 'static + Read + Send + Sync + Unpin,
     {
         let mut form = multipart::Form::default();
 
@@ -134,9 +135,9 @@ pub trait IpfsApi: Backend {
         add: request::Add<'_>,
     ) -> Result<Vec<response::AddResponse>, Self::Error>
     where
-        F: Into<multipart::Form<'static>>,
+        F: Into<multipart::Form<'static>> + Send,
     {
-        let req = self.build_base_request(&add, Some(form.into()))?;
+        let req = self.build_base_request(add, Some(form.into()))?;
         self.request_stream_json(req).try_collect().await
     }
 
@@ -156,7 +157,7 @@ pub trait IpfsApi: Backend {
     ///
     async fn add_path<P>(&self, path: P) -> Result<Vec<response::AddResponse>, Self::Error>
     where
-        P: AsRef<Path>,
+        P: AsRef<Path> + Send,
     {
         let prefix = path.as_ref().parent();
         let mut paths_to_add: Vec<(PathBuf, u64)> = vec![];
@@ -201,7 +202,7 @@ pub trait IpfsApi: Backend {
             }
         }
 
-        let req = self.build_base_request(&request::Add::default(), Some(form))?;
+        let req = self.build_base_request(request::Add::default(), Some(form))?;
 
         self.request_stream_json(req).try_collect().await
     }
@@ -309,7 +310,7 @@ pub trait IpfsApi: Backend {
     ///     .try_concat();
     /// ```
     ///
-    fn block_get(&self, hash: &str) -> Box<dyn Stream<Item = Result<Bytes, Self::Error>> + Unpin> {
+    fn block_get(&self, hash: &str) -> BoxStream<Bytes, Self::Error> {
         impl_stream_api_response! {
             (self, request::BlockGet { hash }, None) => request_stream_bytes
         }
@@ -330,13 +331,49 @@ pub trait IpfsApi: Backend {
     ///
     async fn block_put<R>(&self, data: R) -> Result<response::BlockPutResponse, Self::Error>
     where
-        R: 'static + Read + Send + Sync,
+        R: 'static + Read + Send + Sync + Unpin,
+    {
+        self.block_put_with_options(data, request::BlockPut::default())
+            .await
+    }
+
+    /// Store input as an IPFS block with options.
+    ///
+    /// # Examples
+    ///
+    /// ```no_run
+    /// use ipfs_api::{IpfsApi, IpfsClient};
+    /// use std::io::Cursor;
+    ///
+    /// #[cfg(feature = "with-builder")]
+    /// let options = ipfs_api::request::BlockPut::builder()
+    ///     .mhtype("sha3_384")
+    ///     .mhlen(48)
+    ///     .build();
+    /// #[cfg(not(feature = "with-builder"))]
+    /// let options = ipts_api::request::BlockPut {
+    ///     mhtype: "sha3_384",
+    ///     mhlen: 48,
+    ///     ..Default::default()
+    /// };
+    /// let client = IpfsClient::default();
+    /// let data = Cursor::new("Hello World!");
+    /// let res = client.block_put_with_options(data, options);
+    /// ```
+    ///
+    async fn block_put_with_options<R>(
+        &self,
+        data: R,
+        options: request::BlockPut<'async_trait>,
+    ) -> Result<response::BlockPutResponse, Self::Error>
+    where
+        R: 'static + Read + Send + Sync + Unpin,
     {
         let mut form = multipart::Form::default();
 
         form.add_reader("data", data);
 
-        self.request(request::BlockPut, Some(form)).await
+        self.request(options, Some(form)).await
     }
 
     /// Removes an IPFS block.
@@ -432,7 +469,7 @@ pub trait IpfsApi: Backend {
     ///     .try_concat();
     /// ```
     ///
-    fn cat(&self, path: &str) -> Box<dyn Stream<Item = Result<Bytes, Self::Error>> + Unpin> {
+    fn cat(&self, path: &str) -> BoxStream<Bytes, Self::Error> {
         impl_stream_api_response! {
             (self, request::Cat { path }, None) => request_stream_bytes
         }
@@ -624,7 +661,7 @@ pub trait IpfsApi: Backend {
         data: R,
     ) -> Result<response::ConfigReplaceResponse, Self::Error>
     where
-        R: 'static + Read + Send + Sync,
+        R: 'static + Read + Send + Sync + Unpin,
     {
         let mut form = multipart::Form::default();
 
@@ -662,9 +699,36 @@ pub trait IpfsApi: Backend {
     ///     .try_concat();
     /// ```
     ///
-    fn dag_get(&self, path: &str) -> Box<dyn Stream<Item = Result<Bytes, Self::Error>> + Unpin> {
+    fn dag_get(&self, path: &str) -> BoxStream<Bytes, Self::Error> {
+        self.dag_get_with_options(request::DagGet {
+            path,
+            ..Default::default()
+        })
+    }
+
+    /// Returns information about a dag node in Ipfs with options.
+    ///
+    /// ```no_run
+    /// use futures::TryStreamExt;
+    /// use ipfs_api::{request::{DagCodec, DagGet}, IpfsApi, IpfsClient};
+    ///
+    /// let client = IpfsClient::default();
+    /// let hash = "QmXdNSQx7nbdRvkjGCEQgVjVtVwsHvV8NmV2a8xzQVwuFA";
+    /// #[cfg(feature = "with-builder")]
+    /// let options = DagGet::builder().path(hash).codec(DagCodec::Cbor).build();
+    /// #[cfg(not(feature = "with-builder"))]
+    /// let options = DagGet {
+    ///     path: hash,
+    ///     codec: DagCodec::Cbor,
+    /// };
+    /// client.dag_get_with_options(options)
+    ///     .map_ok(|chunk| chunk.to_vec())
+    ///     .try_concat();
+    /// ```
+    ///
+    fn dag_get_with_options(&self, options: request::DagGet) -> BoxStream<Bytes, Self::Error> {
         impl_stream_api_response! {
-            (self, request::DagGet { path }, None) => request_stream_bytes
+            (self, options, None) => request_stream_bytes
         }
     }
 
@@ -681,13 +745,41 @@ pub trait IpfsApi: Backend {
     ///
     async fn dag_put<R>(&self, data: R) -> Result<response::DagPutResponse, Self::Error>
     where
-        R: 'static + Read + Send + Sync,
+        R: 'static + Read + Send + Sync + Unpin,
+    {
+        self.dag_put_with_options(data, request::DagPut::default())
+            .await
+    }
+
+    /// Add a DAG node to Ipfs with options.
+    ///
+    /// ```no_run
+    /// use ipfs_api::{IpfsApi, IpfsClient};
+    /// use std::io::Cursor;
+    /// use ipfs_api::request::{DagCodec, DagPut};
+    ///
+    /// let client = IpfsClient::default();
+    /// let data = Cursor::new(r#"{ "hello" : "world" }"#);
+    /// let dag_put = DagPut::builder()
+    ///     .input_codec(DagCodec::Json)
+    ///     .pin(false)
+    ///     .build();
+    /// let res = client.dag_put_with_options(data, dag_put);
+    /// ```
+    ///
+    async fn dag_put_with_options<'a, R>(
+        &self,
+        data: R,
+        options: request::DagPut<'a>,
+    ) -> Result<response::DagPutResponse, Self::Error>
+    where
+        R: 'static + Read + Send + Sync + Unpin,
     {
         let mut form = multipart::Form::default();
 
         form.add_reader("object data", data);
 
-        self.request(request::DagPut, Some(form)).await
+        self.request(options, Some(form)).await
     }
 
     // TODO /dag/resolve
@@ -703,10 +795,7 @@ pub trait IpfsApi: Backend {
     /// let res = client.dht_findpeer(peer).try_collect::<Vec<_>>();
     /// ```
     ///
-    fn dht_findpeer(
-        &self,
-        peer: &str,
-    ) -> Box<dyn Stream<Item = Result<response::DhtFindPeerResponse, Self::Error>> + Unpin> {
+    fn dht_findpeer(&self, peer: &str) -> BoxStream<response::DhtFindPeerResponse, Self::Error> {
         impl_stream_api_response! {
             (self, request::DhtFindPeer { peer }, None) => request_stream_json
         }
@@ -723,10 +812,7 @@ pub trait IpfsApi: Backend {
     /// let res = client.dht_findprovs(key).try_collect::<Vec<_>>();
     /// ```
     ///
-    fn dht_findprovs(
-        &self,
-        key: &str,
-    ) -> Box<dyn Stream<Item = Result<response::DhtFindProvsResponse, Self::Error>> + Unpin> {
+    fn dht_findprovs(&self, key: &str) -> BoxStream<response::DhtFindProvsResponse, Self::Error> {
         impl_stream_api_response! {
             (self, request::DhtFindProvs { key }, None) => request_stream_json
         }
@@ -743,10 +829,7 @@ pub trait IpfsApi: Backend {
     /// let res = client.dht_get(key).try_collect::<Vec<_>>();
     /// ```
     ///
-    fn dht_get(
-        &self,
-        key: &str,
-    ) -> Box<dyn Stream<Item = Result<response::DhtGetResponse, Self::Error>> + Unpin> {
+    fn dht_get(&self, key: &str) -> BoxStream<response::DhtGetResponse, Self::Error> {
         impl_stream_api_response! {
             (self, request::DhtGet { key }, None) => request_stream_json
         }
@@ -763,10 +846,7 @@ pub trait IpfsApi: Backend {
     /// let res = client.dht_provide(key).try_collect::<Vec<_>>();
     /// ```
     ///
-    fn dht_provide(
-        &self,
-        key: &str,
-    ) -> Box<dyn Stream<Item = Result<response::DhtProvideResponse, Self::Error>> + Unpin> {
+    fn dht_provide(&self, key: &str) -> BoxStream<response::DhtProvideResponse, Self::Error> {
         impl_stream_api_response! {
             (self, request::DhtProvide { key }, None) => request_stream_json
         }
@@ -782,11 +862,7 @@ pub trait IpfsApi: Backend {
     /// let res = client.dht_put("test", "Hello World!").try_collect::<Vec<_>>();
     /// ```
     ///
-    fn dht_put(
-        &self,
-        key: &str,
-        value: &str,
-    ) -> Box<dyn Stream<Item = Result<response::DhtPutResponse, Self::Error>> + Unpin> {
+    fn dht_put(&self, key: &str, value: &str) -> BoxStream<response::DhtPutResponse, Self::Error> {
         impl_stream_api_response! {
             (self, request::DhtPut { key, value }, None) => request_stream_json
         }
@@ -803,10 +879,7 @@ pub trait IpfsApi: Backend {
     /// let res = client.dht_query(peer).try_collect::<Vec<_>>();
     /// ```
     ///
-    fn dht_query(
-        &self,
-        peer: &str,
-    ) -> Box<dyn Stream<Item = Result<response::DhtQueryResponse, Self::Error>> + Unpin> {
+    fn dht_query(&self, peer: &str) -> BoxStream<response::DhtQueryResponse, Self::Error> {
         impl_stream_api_response! {
             (self, request::DhtQuery { peer }, None) => request_stream_json
         }
@@ -1093,7 +1166,7 @@ pub trait IpfsApi: Backend {
     /// let res = client.files_read("/test/file.json");
     /// ```
     ///
-    fn files_read(&self, path: &str) -> Box<dyn Stream<Item = Result<Bytes, Self::Error>> + Unpin> {
+    fn files_read(&self, path: &str) -> BoxStream<Bytes, Self::Error> {
         self.files_read_with_options(request::FilesRead {
             path,
             ..request::FilesRead::default()
@@ -1124,7 +1197,7 @@ pub trait IpfsApi: Backend {
     fn files_read_with_options(
         &self,
         options: request::FilesRead,
-    ) -> Box<dyn Stream<Item = Result<Bytes, Self::Error>> + Unpin> {
+    ) -> BoxStream<Bytes, Self::Error> {
         impl_stream_api_response! { (self, options, None) => request_stream_bytes }
     }
 
@@ -1236,7 +1309,7 @@ pub trait IpfsApi: Backend {
         data: R,
     ) -> Result<response::FilesWriteResponse, Self::Error>
     where
-        R: 'static + Read + Send + Sync,
+        R: 'static + Read + Send + Sync + Unpin,
     {
         let options = request::FilesWrite {
             path,
@@ -1283,7 +1356,7 @@ pub trait IpfsApi: Backend {
         data: R,
     ) -> Result<response::FilesWriteResponse, Self::Error>
     where
-        R: 'static + Read + Send + Sync,
+        R: 'static + Read + Send + Sync + Unpin,
     {
         let mut form = multipart::Form::default();
 
@@ -1362,9 +1435,7 @@ pub trait IpfsApi: Backend {
     /// let res = client.filestore_dups();
     /// ```
     ///
-    fn filestore_dups(
-        &self,
-    ) -> Box<dyn Stream<Item = Result<response::FilestoreDupsResponse, Self::Error>> + Unpin> {
+    fn filestore_dups(&self) -> BoxStream<response::FilestoreDupsResponse, Self::Error> {
         impl_stream_api_response! {
             (self, request::FilestoreDups, None) => request_stream_json
         }
@@ -1384,7 +1455,7 @@ pub trait IpfsApi: Backend {
     fn filestore_ls(
         &self,
         cid: Option<&str>,
-    ) -> Box<dyn Stream<Item = Result<response::FilestoreLsResponse, Self::Error>> + Unpin> {
+    ) -> BoxStream<response::FilestoreLsResponse, Self::Error> {
         impl_stream_api_response! {
             (self, request::FilestoreLs { cid }, None) => request_stream_json
         }
@@ -1402,8 +1473,7 @@ pub trait IpfsApi: Backend {
     fn filestore_verify(
         &self,
         cid: Option<&str>,
-    ) -> Box<dyn Stream<Item = Result<response::FilestoreVerifyResponse, Self::Error>> + Unpin>
-    {
+    ) -> BoxStream<response::FilestoreVerifyResponse, Self::Error> {
         impl_stream_api_response! {
             (self, request::FilestoreVerify{ cid }, None) => request_stream_json
         }
@@ -1418,7 +1488,7 @@ pub trait IpfsApi: Backend {
     /// let res = client.get("/test/file.json");
     /// ```
     ///
-    fn get(&self, path: &str) -> Box<dyn Stream<Item = Result<Bytes, Self::Error>> + Unpin> {
+    fn get(&self, path: &str) -> BoxStream<Bytes, Self::Error> {
         impl_stream_api_response! {
             (self, request::Get { path }, None) => request_stream_bytes
         }
@@ -1549,11 +1619,11 @@ pub trait IpfsApi: Backend {
     /// let res = client.log_tail();
     /// ```
     ///
-    fn log_tail(&self) -> Box<dyn Stream<Item = Result<String, Self::Error>> + Unpin> {
+    fn log_tail(&self) -> BoxStream<String, Self::Error> {
         impl_stream_api_response! {
             (self, request::LogTail, None) |req| => {
                 self.request_stream(req, |res| {
-                    Self::process_stream_response(res, LineDecoder).map_err(Self::Error::from)
+                    Box::new(Self::process_stream_response(res, LineDecoder).map_err(Self::Error::from))
                 })
             }
         }
@@ -1590,6 +1660,7 @@ pub trait IpfsApi: Backend {
     ///     .path("/ipfs/QmdfTbBqBPQ7VNxZEYEj14VmRuZBkqFbiwReogJgS1zR1n")
     ///     .build()
     /// );
+    /// #[cfg(not(feature = "with-builder"))]
     /// let _ = client.ls_with_options(ipfs_api::request::Ls {
     ///     path: "/ipfs/QmdfTbBqBPQ7VNxZEYEj14VmRuZBkqFbiwReogJgS1zR1n",
     ///     // Example options for fast listing
@@ -1602,7 +1673,7 @@ pub trait IpfsApi: Backend {
     fn ls_with_options(
         &self,
         options: request::Ls<'_>,
-    ) -> Box<dyn Stream<Item = Result<response::LsResponse, Self::Error>> + Unpin> {
+    ) -> BoxStream<response::LsResponse, Self::Error> {
         impl_stream_api_response! {
             (self, options, None) => request_stream_json
         }
@@ -1685,7 +1756,7 @@ pub trait IpfsApi: Backend {
     /// let res = client.object_data("/ipfs/QmVrLsEDn27sScp3k23sgZNefVTjSAL3wpgW1iWPi4MgoY");
     /// ```
     ///
-    fn object_data(&self, key: &str) -> Box<dyn Stream<Item = Result<Bytes, Self::Error>> + Unpin> {
+    fn object_data(&self, key: &str) -> BoxStream<Bytes, Self::Error> {
         impl_stream_api_response! {
             (self, request::ObjectData { key }, None) => request_stream_bytes
         }
@@ -1887,7 +1958,7 @@ pub trait IpfsApi: Backend {
         &self,
         peer: &str,
         count: Option<i32>,
-    ) -> Box<dyn Stream<Item = Result<response::PingResponse, Self::Error>> + Unpin> {
+    ) -> BoxStream<response::PingResponse, Self::Error> {
         impl_stream_api_response! {
             (self, request::Ping { peer, count }, None) => request_stream_json
         }
@@ -1912,33 +1983,64 @@ pub trait IpfsApi: Backend {
     /// use ipfs_api::{IpfsApi, IpfsClient};
     ///
     /// let client = IpfsClient::default();
-    /// let res = client.pubsub_peers(None);
+    /// let res = client.pubsub_peers(Option::<String>::None);
     /// let res = client.pubsub_peers(Some("feed"));
     /// ```
     ///
-    async fn pubsub_peers(
+    async fn pubsub_peers<T>(
         &self,
-        topic: Option<&str>,
-    ) -> Result<response::PubsubPeersResponse, Self::Error> {
-        self.request(request::PubsubPeers { topic }, None).await
+        topic: Option<T>,
+    ) -> Result<response::PubsubPeersResponse, Self::Error>
+    where
+        T: AsRef<[u8]> + Send + Sync,
+    {
+        match topic {
+            Some(topic) => {
+                self.request(
+                    request::PubsubPeers {
+                        topic: Some(topic.as_ref()),
+                    },
+                    None,
+                )
+                .await
+            }
+            None => {
+                self.request(request::PubsubPeers { topic: None }, None)
+                    .await
+            }
+        }
     }
 
     /// Publish a message to a topic.
     ///
     /// ```no_run
     /// use ipfs_api::{IpfsApi, IpfsClient};
+    /// use std::io::Cursor;
     ///
     /// let client = IpfsClient::default();
-    /// let res = client.pubsub_pub("feed", "Hello World!");
+    /// let res = client.pubsub_pub("feed", Cursor::new("Hello World!"));
     /// ```
     ///
-    async fn pubsub_pub(
+    async fn pubsub_pub<T, R>(
         &self,
-        topic: &str,
-        payload: &str,
-    ) -> Result<response::PubsubPubResponse, Self::Error> {
-        self.request_empty(request::PubsubPub { topic, payload }, None)
-            .await
+        topic: T,
+        data: R,
+    ) -> Result<response::PubsubPubResponse, Self::Error>
+    where
+        T: AsRef<[u8]> + Send + Sync,
+        R: 'static + Read + Send + Sync + Unpin,
+    {
+        let mut form = multipart::Form::default();
+
+        form.add_reader("data", data);
+
+        self.request_empty(
+            request::PubsubPub {
+                topic: topic.as_ref(),
+            },
+            Some(form),
+        )
+        .await
     }
 
     /// Subscribes to a pubsub topic.
@@ -1947,17 +2049,16 @@ pub trait IpfsApi: Backend {
     /// use ipfs_api::{IpfsApi, IpfsClient};
     ///
     /// let client = IpfsClient::default();
-    /// let res = client.pubsub_sub("feed", false);
-    /// let res = client.pubsub_sub("feed", true);
+    /// let res = client.pubsub_sub("feed");
+    /// let res = client.pubsub_sub("feed");
     /// ```
     ///
-    fn pubsub_sub(
-        &self,
-        topic: &str,
-        discover: bool,
-    ) -> Box<dyn Stream<Item = Result<response::PubsubSubResponse, Self::Error>> + Unpin> {
+    fn pubsub_sub<T>(&self, topic: T) -> BoxStream<response::PubsubSubResponse, Self::Error>
+    where
+        T: AsRef<[u8]>,
+    {
         impl_stream_api_response! {
-            (self, request::PubsubSub { topic, discover }, None) => request_stream_json
+            (self, request::PubsubSub { topic: topic.as_ref() }, None) => request_stream_json
         }
     }
 
@@ -1970,9 +2071,7 @@ pub trait IpfsApi: Backend {
     /// let res = client.refs_local();
     /// ```
     ///
-    fn refs_local(
-        &self,
-    ) -> Box<dyn Stream<Item = Result<response::RefsLocalResponse, Self::Error>> + Unpin> {
+    fn refs_local(&self) -> BoxStream<response::RefsLocalResponse, Self::Error> {
         impl_stream_api_response! {
             (self, request::RefsLocal, None) => request_stream_json
         }
@@ -2094,7 +2193,7 @@ pub trait IpfsApi: Backend {
     ///
     async fn tar_add<R>(&self, data: R) -> Result<response::TarAddResponse, Self::Error>
     where
-        R: 'static + Read + Send + Sync,
+        R: 'static + Read + Send + Sync + Unpin,
     {
         let mut form = multipart::Form::default();
 
@@ -2112,7 +2211,7 @@ pub trait IpfsApi: Backend {
     /// let res = client.tar_cat("/ipfs/QmVrLsEDn27sScp3k23sgZNefVTjSAL3wpgW1iWPi4MgoY");
     /// ```
     ///
-    fn tar_cat(&self, path: &str) -> Box<dyn Stream<Item = Result<Bytes, Self::Error>> + Unpin> {
+    fn tar_cat(&self, path: &str) -> BoxStream<Bytes, Self::Error> {
         impl_stream_api_response! {
             (self, request::TarCat { path }, None) => request_stream_bytes
         }
